/*
 * Copyright (c) 2004, 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sockmapimpl.c,v 1.15 2008/02/10 04:54:39 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "sm/string.h"
#include "sockmap.h"

/* --------- socket map basic implementation ------ */

#if MTA_USE_PTHREADS 
# define SOCKMAP_LOCK_INIT(map) 					\
	pthread_rwlock_init(&(map)->sockmap_lock, NULL)
# define SOCKMAP_LOCK_DESTROY(map)			                \
	pthread_rwlock_destroy(&(map)->sockmap_lock)
# define SOCKMAP_LOCK(map) do {						\
		int r = pthread_rwlock_wrlock(&(map)->sockmap_lock);	\
		if (r != 0)						\
			return sm_error_perm(SM_EM_MAP, r);		\
	} while(0)
# define SOCKMAP_UNLOCK(map) do {					\
		int r = pthread_rwlock_unlock(&(map)->sockmap_lock);	\
		if (r != 0)						\
			return sm_error_perm(SM_EM_MAP, r);		\
	} while(0)
# define SOCKMAP_INST_LOCK_INIT(map, i)				\
	pthread_rwlock_init(&(map)->sockmap_inst[i].rwlock, NULL)
# define SOCKMAP_INST_LOCK_DESTROY(map, i)			\
	pthread_rwlock_destroy(&(map)->sockmap_inst[i].rwlock)
# define SOCKMAP_INST_LOCK(map, i) do {					\
		int r = pthread_rwlock_wrlock(&(map)->sockmap_inst[i].rwlock); \
		if (r != 0)						\
			return sm_error_perm(SM_EM_MAP, r);		\
	} while(0)
# define SOCKMAP_INST_UNLOCK(map, i) do {				\
		int r = pthread_rwlock_unlock(&(map)->sockmap_inst[i].rwlock); \
		if (r != 0)						\
			return sm_error_perm(SM_EM_MAP, r);		\
	} while(0)
#else
# define SOCKMAP_LOCK_INIT(map)	            0
# define SOCKMAP_LOCK_DESTROY(map)	    0
# define SOCKMAP_LOCK(map)
# define SOCKMAP_UNLOCK(map)
# define SOCKMAP_INST_LOCK_INIT(map, i)	    0
# define SOCKMAP_INST_LOCK_DESTROY(map, i)  0
# define SOCKMAP_INST_LOCK(map, i)
# define SOCKMAP_INST_UNLOCK(map, i)
#endif

static sm_ret_T
sockmap_init_all_instances(sm_sockmap_P db)
{
	uint i;
	db->sockmap_inst = sm_zalloc(db->sockmap_max
				     * sizeof(db->sockmap_inst[0]));
	if (!db->sockmap_inst)
		return sm_error_perm(SM_EM_MAP, ENOMEM);
	
	for (i = 0; i < db->sockmap_max; i++)  
		SOCKMAP_INST_LOCK_INIT(db, i);
	return SM_SUCCESS;
}
	
static sm_ret_T
sockmap_close_instance(sm_sockmap_P db, uint i)
{
	sm_ret_T ret = SM_SUCCESS;
	struct sm_sockmap_instance_S *smp = db->sockmap_inst + i;
	if (smp->fp) {
		ret = sm_io_close(smp->fp, SM_IO_CF_NONE);
		smp->fp = NULL;
		smp->busy = 0;
	}
	return ret;
}

static sm_ret_T
sockmap_open_instance(sm_sockmap_P db, uint i)
{
	sm_ret_T ret = SM_SUCCESS;
	struct sm_sockmap_instance_S *smp = db->sockmap_inst + i;
	int fd;

	if (smp->fp)
		return SM_SUCCESS;
			
	if (db->sockmap_path != NULL)
		ret = unix_client_connect(db->sockmap_path, &fd);
	else
		ret = net_client_connectipv4(db->sockmap_ipv4,
					     db->sockmap_port, &fd);
	if (sm_is_err(ret))
		goto fail;
	ret = sm_io_open(SmStStdiofd, (void *) &fd, SM_IO_RDWR,
			 &smp->fp, SM_IO_WHAT_END);
	if (sm_is_err(ret))
		goto fail;
	ret = sm_io_setinfo(smp->fp, SM_IO_DOUBLE, NULL);
	if (sm_is_err(ret))
		goto fail;
	ret = sm_io_setinfo(smp->fp, SM_IO_WHAT_TIMEOUT,
			    (void *)&db->sockmap_tmout);
	if (sm_is_err(ret)) {
	  fail:
		sockmap_close_instance(db, i);
	}
	return ret;
}

static sm_ret_T
sockmap_find_free_instance0(sm_sockmap_P db, uint *rind)
{
	int i;
	int busy_ind = -1;
	int free_ind = -1;
	
	for (i = 0; i < db->sockmap_max; i++) {
		if (db->sockmap_inst[i].fp == NULL) {
			if (free_ind == -1)
				free_ind = i;
		} else if (!db->sockmap_inst[i].busy) {
			db->sockmap_inst[i].busy = 1;
			*rind = i;
			return SM_SUCCESS;
		} else if (busy_ind != -1)
			busy_ind = i;
	}
	if (free_ind != -1) {
		sm_ret_T ret = sockmap_open_instance(db, free_ind);
		if (!sm_is_error(ret))
			*rind = free_ind;
		return ret;
	}
	if (busy_ind != 0) {
		/* Degrading to serialized access */
		*rind = busy_ind;
		return SM_SUCCESS;
	}
	/* Giving up... */
	return sm_error_perm(SM_EM_MAP, EAGAIN);
}

static sm_ret_T
sockmap_find_free_instance(sm_sockmap_P db, uint *rind)
{
	sm_ret_T ret;
	int r;
	
	SOCKMAP_LOCK(db);
	ret = sockmap_find_free_instance0(db, rind);
	SOCKMAP_UNLOCK(db);
	return ret;
}

static sm_ret_T
sockmap_release_instance(sm_sockmap_P db, uint i)
{
	db->sockmap_inst[i].busy = 0;
	return SM_SUCCESS;
}

/* FIXME: Return code? */
static void
sockmap_close_all_instances(sm_sockmap_P db, int destr)
{
	uint i;

	if (!db->sockmap_inst)
		return;
	for (i = 0; i < db->sockmap_max; i++) {
		sockmap_close_instance(db, i);
		if (destr)
			SOCKMAP_INST_LOCK_DESTROY(db, i);
	}
}

static sm_ret_T
sockmap_instance_lookup(sm_sockmap_P map, uint i, sm_str_P key, sm_str_P data)
{
	sm_ret_T ret;
	uint len, replylen;
	int c;
	ssize_t recvlen;
	char *value, *status;
	sm_file_T *fp;
	char statbuf[16];
	struct sm_sockmap_instance_S *smp = map->sockmap_inst + i;
	
	fp = smp->fp;
	if (fp == NULL)
	{
		ret = sockmap_open(map);
		if (sm_is_err(ret))
			return ret;
	}
	len = strlen(map->sockmap_name) + 1 + sm_str_getlen(key);
	SM_ASSERT(len > strlen(map->sockmap_name));
	SM_ASSERT(len > sm_str_getlen(key));
	if ((sm_io_fprintf(fp, "%u:%s %S,", len, map->sockmap_name, key)
	     == SM_IO_EOF) ||
	    (sm_io_flush(fp) != 0) || (sm_io_error(fp) != 0))
	{
		ret = sm_error_temp(SM_EM_MAP, errno);
		goto errcl;
	}

	if (sm_io_fscanf(fp, "%9u", &replylen) != 1) {
		ret = sm_error_temp(SM_EM_MAP, errno);
		goto errcl;
	}
	if (replylen > SOCKETMAP_MAXL) {
		ret = sm_error_temp(SM_EM_MAP, SM_E_PR_ERR);
		goto errcl;
	}
	if (sm_io_getc(fp) != ':') {
		ret = sm_error_temp(SM_EM_MAP, SM_E_PR_ERR);
		goto error;
	}

	len = 0;
	sm_memzero(statbuf, sizeof(statbuf));
	if (replylen >= sm_str_getsize(data)) {
		/* not enough space: read status and discard rest */
		while (replylen-- > 0 && (c = sm_io_getc(fp)) != SM_IO_EOF) {
			if (len < sizeof(statbuf) - 1)
				statbuf[len++] = c;
		}
	}
	else {
		/* read status first */
		c = '\0';
		while (replylen-- > 0 && (c = sm_io_getc(fp)) != SM_IO_EOF && c != ' ')
		{
			if (len < sizeof(statbuf) - 1)
				statbuf[len++] = c;
		}
		if (replylen > 0 && c == ' ') {
			ret = sm_io_read(fp, sm_str_data(data), replylen, &recvlen);
			if (sm_is_err(ret))
				goto errcl;
			else if (recvlen < replylen) {
				ret = sm_error_temp(SM_EM_MAP, errno);
				goto errcl;
			}
			SM_STR_SETLEN(data, recvlen);
		}
	}
	if (sm_io_getc(fp) != ',') {
		ret = sm_error_temp(SM_EM_MAP, SM_E_PR_ERR);
		goto errcl;
	}
	statbuf[sizeof(statbuf) - 1] = '\0';
	status = statbuf;

	value = strchr(status, ' ');
	if (value != NULL) {
		*value = '\0';
		value++;
	}
	if (sm_streq(status, "OK")) {
		ret = SM_SUCCESS;
	}
	else if (sm_streq(status, "NOTFOUND")) {
		return sm_error_perm(SM_EM_MAP, SM_E_NOTFOUND);
	}
	else {
		if (sm_streq(status, "TEMP") || sm_streq(status, "TIMEOUT"))
			ret = sm_error_temp(SM_EM_MAP, SM_E_TEMPMAP);
		else if (sm_streq(status, "PERM"))
			ret = sm_error_perm(SM_EM_MAP, SM_E_PERMMAP);
		else
			ret = sm_error_temp(SM_EM_MAP, SM_E_PR_ERR);
	}

	return ret;

  errcl:
	sockmap_close_instance(map, i);

  error:
	return ret;
}


/*
**  SOCKMAP_FREE -- free socket map context
**
**	Parameters:
**		db -- pointer to socket map
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sockmap_free(sm_sockmap_P db)
{
	if (db != NULL) {
		if (db->sockmap_inst)
			sm_free(db->sockmap_inst);
		sm_free_size(db, sizeof(*db));
	}
	return SM_SUCCESS;
}

/*
**  SOCKMAP_NEW -- allocate socket map context
**
**	Parameters:
**		pdb -- pointer to pointer to socket map
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sockmap_new(sm_sockmap_P *pdb)
{
	sm_sockmap_P db;
	int rc;
	
	SM_REQUIRE(pdb != NULL);
	*pdb = NULL;
	db = sm_zalloc(sizeof(*db));
	if (db == NULL)
		return sm_error_perm(SM_EM_MAP, ENOMEM);
	rc = SOCKMAP_LOCK_INIT(db);
	if (rc) {
		sm_free(db);
		return sm_error_perm(SM_EM_MAP, rc);
	}

        /* default for socket map timeout */
	db->sockmap_tmout = 5;
	db->sockmap_min = db->sockmap_max = 1;
	*pdb = db;
	return SM_SUCCESS;
}

/*
**  SOCKMAP_DESTROY -- destroy socket map
**
**	Parameters:
**		pdb -- pointer to pointer to socket map
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sockmap_destroy(sm_sockmap_P *pdb)
{
	sm_ret_T ret;

	SM_REQUIRE(pdb != NULL);
	ret = SM_SUCCESS;
	if (*pdb != NULL) {
		sockmap_close_all_instances(*pdb, 1);
		SOCKMAP_LOCK_DESTROY(*pdb);
		sockmap_free(*pdb);
		*pdb = NULL;
	}
	return ret;
}

/*
**  SOCKMAP_CLOSE -- close socket map
**
**	Parameters:
**		db -- pointer to socket map
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sockmap_close(sm_sockmap_P db)
{
	sm_ret_T ret;

	ret = SM_SUCCESS;
	sockmap_close_all_instances(db, 0);
	return ret;
}

/*
**  SOCKMAP_OPEN -- open socket map
**
**	Parameters:
**		pdb -- pointer to pointer to socket map
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sockmap_open(sm_sockmap_P db)
{
	sm_ret_T ret;
	uint i, j;
	
	SM_REQUIRE(db != NULL);
	if (!db->sockmap_inst) {
		ret = sockmap_init_all_instances(db);
		if (sm_is_error(ret))
			return ret;
	}

	for (i = 0, j = 0; i < db->sockmap_max && j < db->sockmap_min; i++) {
		if (db->sockmap_inst[i].fp == NULL) {
			ret = sockmap_open_instance(db, i);
			if (sm_is_error(ret))
				break;
		}
		j++;
	}

	if (j > 0 && ret)
		/* Try to continue anyway */
		ret = SM_SUCCESS;

	if (sm_is_error(ret)) {
		/* cleanup? */
		sockmap_close(db);
	}

	return ret;
}

/*
**  SOCKMAP_LOOKUP -- lookup a key in SOCKMAP, return data if found
**
**	Parameters:
**		map -- map context
**		key -- key
**		data -- data (output)
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sockmap_lookup(sm_sockmap_P map, sm_str_P key, sm_str_P data)
{
	sm_ret_T ret;
	uint ninst;
	
	ret = sockmap_find_free_instance(map, &ninst);
	if (sm_is_error(ret))
		return ret;
	SOCKMAP_INST_LOCK(map, ninst);
	ret = sockmap_instance_lookup(map, ninst, key, data);
	sockmap_release_instance(map, ninst);
	SOCKMAP_INST_UNLOCK(map, ninst);
	return ret;
}

