/*
 * Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-sockmap-0.c,v 1.10 2008/02/10 06:17:29 ca Exp $")

#include "sm/error.h"
#include "sm/heap.h"
#include "sm/memops.h"
#include "sm/sysexits.h"
#include "sm/test.h"
#include "sm/maps.h"
#include "sm/mapc.h"
#include "sm/map.h"
#include "sm/mapclasses.h"
#include "sm/net.h"
#include <stdio.h>
#include "sm/io.h"

int Verbose = 0;

#define MAPC_TYPE	"socket"
#define MAPC_NAME	"name"

sm_map_P map;
size_t test_run_count;

static pthread_cond_t signal_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t signal_mutex = PTHREAD_MUTEX_INITIALIZER;

struct test_tab {
	const char *key;
	int (*is_ok)(sm_ret_T, char **);
	pthread_t thr;
};

int
test_success(sm_ret_T ret, char **pret)
{
	return sm_is_success(ret);
}

int
test_notfound(sm_ret_T ret, char **pret)
{
	if (!sm_is_err(ret))
		*pret = "unexpected successful return";
	if (SM_MAP_NOTFOUND != ret) {
		*pret = smerr2txt(ret);
		return 0;
	}
	return 1;
}

int
test_err(sm_ret_T ret, char **pret)
{
	if (!sm_is_err(ret)) {
		*pret = "unexpected successful return";
		return 0;
	}
	return 1;
}

struct test_tab test_tab[] = {
	{ "bastiaan.bakker@example.com", test_success },
	{ "wolter.eldering@example.com", test_success },
	{ "no@example.com", test_notfound },
	{ "TIMEOUT", test_err },
	{ "wolter.eldering@example.com", test_success },
	{ "bastiaan.bakker@example.com", test_success },
	{ NULL }
};

#define incr_test_run_count() test_run_count++

void
decr_test_run_count()
{
	if (--test_run_count == 0) {
		pthread_mutex_lock(&signal_mutex);
		pthread_cond_signal(&signal_cond);
		pthread_mutex_unlock(&signal_mutex);
	}
}

void *
test_thr(void *data)
{
	const char *arg = (const char*) data;
	sm_str_P lhs, rhs;
	sm_ret_T ret;
	lhs = rhs = NULL;

	incr_test_run_count();
	if (Verbose)
		fprintf(stderr, "%s: BEGIN\n", arg);
	lhs = sm_str_new(NULL, 256, 1024);
	SM_ASSERT(lhs != NULL);

	rhs = sm_str_new(NULL, 256, 1024);
	SM_ASSERT(rhs != NULL);

	sm_str_clr(rhs);
	sm_str_clr(lhs);
	sm_str_scat(lhs, arg);
	ret = sm_map_lookup(map, SMMAP_FL_NONE, lhs, rhs);
	if (Verbose)
		fprintf(stderr, "%s: FINISHED: %x = %s\n",
			arg, ret, ret == 0 ? "OK" : smerr2txt(ret));
  error:
	SM_STR_FREE(lhs);
	SM_STR_FREE(rhs);
	decr_test_run_count();
	return (void*) ret;
}


static void
testh(const char *name, const char *path, ipv4_T ipv4, short port)
{
	sm_ret_T ret;
	sm_maps_P maps;
	sm_cstr_P mtype, mname;
	int i;
	
	maps = NULL;
	mtype = mname = NULL;

	ret = sm_maps_init(&maps);
	SM_TEST(maps != NULL);
	if (maps == NULL)
		return;
	SM_TEST(sm_is_success(ret));

	mtype = sm_cstr_scpyn0((const uchar *)MAPC_TYPE, strlen(MAPC_TYPE));
	SM_TEST_E(mtype != NULL);

	mname = sm_cstr_scpyn0((const uchar *)MAPC_NAME, strlen(MAPC_NAME));
	SM_TEST_E(mname != NULL);

	ret = sm_sockmap_class_create(maps);
	SM_TEST(sm_is_success(ret));

	map = NULL;
	ret = sm_map_open(maps, mname, mtype, 0, name, SMAP_MODE_RDWR, &map,
			  SMPO_SOCKPATH, path,
			  SMPO_PORT, port, SMPO_IPV4, ipv4,
			  SMPO_MIN_OCONN, 1, SMPO_MAX_OCONN, 10,
//			  SMPO_TMOUT, 3600,
			  SMPO_END);
	SM_TEST_E(sm_is_success(ret));

	for (i = 0; test_tab[i].key; i++) 
		pthread_create(&test_tab[i].thr, NULL, test_thr,
			       (void*)test_tab[i].key);

	pthread_mutex_lock(&signal_mutex);
	pthread_cond_wait(&signal_cond, &signal_mutex);
	pthread_mutex_unlock(&signal_mutex);

	for (i = 0; test_tab[i].key; i++) {
		void *val;
		sm_ret_T ret;
		char *text = NULL; 
		pthread_join(test_tab[i].thr, &val);
		ret = val;
		if (test_tab[i].is_ok(ret, &text) == 0) {
			fprintf(stderr, "%d: %s: FAIL", i, test_tab[i].key);
			if (text)
				fprintf(stderr, " %s", text);
			fprintf(stderr, "\n");
		} else
			fprintf(stderr, "%d: %s: OK\n", i, test_tab[i].key);
	}
	
	ret = sm_map_close(map, 0);
	SM_TEST(sm_is_success(ret));

	ret = sm_maps_term(maps);
	SM_TEST(sm_is_success(ret));
	SM_CSTR_FREE(mtype);
	SM_CSTR_FREE(mname);
	return;

  error:
	sm_maps_term(maps);
}

#define T_SM_NAME "virtuser"
#define T_SM_PATH NULL
#define T_SM_PORT 1289
#define T_SM_IP 0x7F000001

static sm_ret_T
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options]\n"
		"-a ip     IPv4 address [%#x]\n"
		"-N name   name of map [%s]\n"
		"-P path   path [%s]\n"
		"-p port   port [%d]\n"
		"-v\n"      
		, prg
		, T_SM_IP
		, T_SM_NAME
		, T_SM_PATH
		, T_SM_PORT
		);
	return EX_USAGE;
}

int
main(int argc, char *argv[])
{
	char *path, *name, *prg;
	ipv4_T ipv4;
	short port;
	int c;

	prg = argv[0];
	path = NULL;
	name = "virtuser";
	port = 1289;
	ipv4 = ntohl(0x7F000001);
	while ((c = getopt(argc, argv, "a:N:P:p:H:s:V")) != -1)
	{
		switch (c)
		{
		  case 'a':
			ipv4 = (ipv4_T) atol(optarg);
			break;
		  case 'N':
			name = strdup(optarg);
			if (name == NULL)
				return 1;
			break;
		  case 'P':
			path = strdup(optarg);
			if (path == NULL)
				return 1;
			break;
		  case 'p':
			port = atoi(optarg);
			break;
		  case 'V':
			Verbose++;
			break;
		  default:
			usage(prg);
			return EX_USAGE;
		}
	}
	sm_test_begin(argc, argv, "test socket map 0");
	testh(name, path, ipv4, port);
	return sm_test_end();
}
